"""Store date-related regex patterns and constants."""

import logging

from strenum import StrEnum

from monopoly.enums import RegexEnum

logger = logging.getLogger(__name__)


# pylint: disable=line-too-long
class DateFormats(StrEnum):
    """Case-insensitive list of common ISO 8601 date formats."""

    D = r"(1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    DD = r"(01|02|03|04|05|06|07|08|09|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    M = r"(1|2|3|4|5|6|7|8|9|10|11|12)"
    MM = r"(01|02|03|04|05|06|07|08|09|10|11|12)"
    MMM = r"(?i:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"
    MMMM = r"(?i:January|February|March|April|May|June|July|August|September|October|November|December)"
    YY = r"([2-5][0-9]\b)"
    YYYY = r"(20\d{2}\b)"


class ISO8601(RegexEnum):
    D_MMM = rf"\b({DateFormats.D}[\/\-\s.]{DateFormats.MMM})"
    D_MMM_RELAXED = D_MMM.replace(r"[\/\-\s.]", r"(?:[\/\-\s.]|)")
    DD_MM = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MM})"
    DD_MM_YY = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MM}[\/\-\s.]{DateFormats.YY})"
    DD_MM_YYYY = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MM}[\/\-\s.]{DateFormats.YYYY})"
    DD_MMM = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MMM})"
    DD_MMM_RELAXED = DD_MMM.replace(r"[\/\-\s.]", r"(?:[\/\-\s.]|)")
    DD_MMM_YY = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MMM}[\/\-\s.]{DateFormats.YY})"
    DD_MMM_YYYY = rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MMM}[,\s]{{1,2}}{DateFormats.YYYY})"
    MM_DD = rf"\b({DateFormats.MM}[\/\-\s.]{DateFormats.DD})"
    MM_DD_YY = rf"\b({DateFormats.MM}[\/\-\s.]{DateFormats.DD}[\/\-\s.]{DateFormats.YY})"
    MM_DD_YYYY = rf"\b({DateFormats.MM}[\/\-\s.]{DateFormats.DD}[\/\-\s.]{DateFormats.YYYY})"
    MMMM_DD_YYYY = rf"\b({DateFormats.MMMM}\s{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"
    MMM_DD = rf"\b({DateFormats.MMM}[\/\-\s.]{DateFormats.DD})"
    MMM_DD_YYYY = rf"\b({DateFormats.MMM}[\/\-\s.]{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"



import json
import re
from collections.abc import Mapping
from typing import Any

from pydantic import ConfigDict, Field, field_validator, model_validator
from pydantic.dataclasses import dataclass
from pydantic_core import ArgsKwargs

from monopoly.constants import Columns


# ruff: noqa: N805
class TransactionGroupDict(Mapping):
    """
    Helper class that holds the group dict from a match.

    This class acts as a convenient wrapper to allow unpacking into the
    `Transaction` class.

    A transaction should at minimum always have a description and amount

    In cases where we parse a previous balance line, the date might not exist
    e.g.
    `          LAST MONTH'S BALANCE         6,321`
    `01 OCT    ValueVille                  123.12`
    """

    def __init__(
        self,
        description: str,
        amount: str,
        transaction_date: str | None = None,
        polarity: str | None = None,
        **_,
    ):
        self.transaction_date = transaction_date
        self.amount = amount
        self.description = description
        self.polarity = polarity

    def __getitem__(self, x):
        return self.__dict__[x]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def asdict(self):
        return self.__dict__

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return (
            "<TransactionGroupDict object; "
            f"{Columns.TRANSACTION_DATE}={self.transaction_date}, "
            f"{Columns.AMOUNT}={self.amount}, "
            f"{Columns.DESCRIPTION}={self.description}>"
        )


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class TransactionMatch:
    groupdict: TransactionGroupDict
    match: re.Match
    page_number: int

    def span(self):
        return self.match.span()


@dataclass
class Transaction:
    """
    Hold transaction data, validates the data, and performs various coercions.

    This includes removing whitespaces and commas, reassigning 'transaction_date' to 'date'
    """

    description: str
    amount: float
    date: str = Field(alias="transaction_date")
    polarity: str | None = None
    # avoid storing config logic, since the Transaction object is used to create
    # a single unique hash which should not change
    auto_polarity: bool = Field(default=True, init=True, repr=False)

    def as_raw_dict(self, *_, show_polarity=False):
        """Return stringified dictionary version of the transaction."""
        items = {
            Columns.DATE.value: self.date,
            Columns.DESCRIPTION.value: self.description,
            Columns.AMOUNT.value: str(self.amount),
        }
        if show_polarity:
            items[Columns.POLARITY] = self.polarity
        return items

    @field_validator("description", mode="after")
    def remove_extra_whitespace(cls, value: str) -> str:
        return " ".join(value.split())

    @field_validator(Columns.AMOUNT, mode="before")
    def prepare_amount_for_float_coercion(cls, amount: str) -> str:
        """
        Replace commas, whitespaces, apostrophes and parentheses for string representation of floats.

        1'234.00 -> 1234.00
        1,234.00 -> 1234.00
        (-10.00) -> -10.00
        (-1.56 ) -> -1.56.
        """
        if isinstance(amount, str):
            return re.sub(r"[^\d\.\-]", "", amount)
        return amount

    # pylint: disable=bad-classmethod-argument
    @model_validator(mode="before")
    def treat_parenthesis_enclosure_as_credit(self: ArgsKwargs | Any) -> "ArgsKwargs":
        """Treat amounts enclosed by parentheses (e.g. cashback) as a credit entry."""
        if self.kwargs:
            amount: str = self.kwargs[Columns.AMOUNT]
            if isinstance(amount, str) and amount.startswith("(") and amount.endswith(")"):
                self.kwargs[Columns.POLARITY] = "CR"
        return self

    @model_validator(mode="after")
    def convert_credit_amount_to_negative(self: "Transaction") -> "Transaction":
        """Convert transactions with a polarity of "CR" or "+" to positive."""
        # avoid negative zero
        if self.amount == 0:
            return self

        if not self.auto_polarity:
            return self

        if self.polarity in ("CR", "+"):
            self.amount = abs(self.amount)

        else:
            self.amount = -abs(self.amount)
        return self

    def __str__(self):
        return json.dumps(self.as_raw_dict(show_polarity=True))



from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import pytest

from monopoly.pipeline import Pipeline
from monopoly.statements import BaseStatement, Transaction


@pytest.fixture
def mock_generate_name():
    with patch("monopoly.pipeline.generate_name", return_value="test_file.csv") as mock:
        yield mock


@pytest.fixture
def mock_file_system():
    with patch("builtins.open", new_callable=MagicMock) as mock_open:
        with patch("csv.writer", autospec=True) as mock_writer:
            mock_instance = mock_writer.return_value
            mock_instance.writerow = MagicMock()
            yield mock_open, mock_instance


@pytest.mark.usefixtures("mock_generate_name")
def test_load(
    credit_statement: BaseStatement,
    mock_file_system,
):
    mock_open, mock_csv_writer = mock_file_system

    transactions = [
        Transaction(transaction_date="2023-01-01", description="foo", amount=100.0),
        Transaction(transaction_date="2023-01-01", description="bar", amount=123.12),
    ]

    credit_statement.statement_date = datetime(2023, 1, 1)

    output_path = Pipeline.load(
        transactions=transactions,
        statement=credit_statement,
        output_directory=Path("/output_directory"),
        preserve_filename=False,
    )

    expected = Path("/output_directory/test_file.csv")
    mock_open.assert_called_once_with(expected, mode="w", encoding="utf8")
    expected_calls = [
        call.writerow(["date", "description", "amount"]),
        call.writerow(["2023-01-01", "foo", -100.0]),
        call.writerow(["2023-01-01", "bar", -123.12]),
    ]

    mock_csv_writer.writerow.assert_has_calls(expected_calls)
    assert output_path == expected



from monopoly.statements.base import BaseStatement, MatchContext


def test_multiline_polarity_detected(statement: BaseStatement):
    statement.pages[0].lines = [
        "24.02.25  PAYMENT RECEIVED - THANK YOU            79.99",
        "                                                  CR",
    ]

    context = MatchContext(
        lines=statement.pages[0].lines,
        line=statement.pages[0].lines[0],
        description="PAYMENT RECEIVED - THANK YOU",
        idx=0,
    )

    result = statement.get_multiline_polarity(context)
    assert result == "CR"



import csv
import logging
import re
from datetime import datetime
from pathlib import Path

from pydantic import SecretStr

from monopoly.constants.date import DateFormats
from monopoly.generic import GenericBank, GenericStatementHandler
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfParser
from monopoly.statements import BaseStatement, Transaction
from monopoly.write import generate_name

logger = logging.getLogger(__name__)

START_OF_YEAR_MONTHS = (1, 2)
YEAR_CUTOFF_MONTH = 2


class Pipeline:
    """Handles extract, transform and load (ETL) logic for bank statements."""

    def __init__(
        self,
        parser: PdfParser,
        passwords: list[SecretStr] | None = None,
    ):
        self.passwords = passwords
        self.handler = self.create_handler(parser)

    @staticmethod
    def create_handler(parser: PdfParser) -> StatementHandler:
        if issubclass(parser.bank, GenericBank):
            logger.debug("Using generic statement handler")
            return GenericStatementHandler(parser)
        logger.debug("Using statement handler with bank: %s", parser.bank.__name__)
        return StatementHandler(parser)

    def extract(self, *_, safety_check=True) -> BaseStatement:
        """
        Extract transactions from the statement.

        Perform a safety check to make sure that total transactions add up.
        """
        statement = self.handler.statement

        if not statement.transactions:
            msg = "No transactions found - statement extraction failed"
            raise ValueError(msg)

        logger.debug("%s transactions found", len(statement.transactions))

        if not statement.statement_date:
            msg = "No statement date found"
            raise ValueError(msg)

        if safety_check and statement.config.safety_check:
            statement.perform_safety_check()

        return statement

    @staticmethod
    def transform(statement: BaseStatement) -> list[Transaction]:
        logger.debug("Running transformation functions on DataFrame")

        transactions = statement.transactions
        statement_date = statement.statement_date
        date_order = statement.config.transaction_date_order
        date_format = statement.config.transaction_date_format

        def convert_date(tx: Transaction) -> str:
            """
            Convert date to ISO 8601 format with cross-year logic.

            Applies the following logic:
            - If the transaction date does not include a year, append the year from the statement date.
            - Attempts to parse the date using a specified format (for performance).
            - Falls back to a flexible date parser if format-based parsing fails.
            - Applies cross-year adjustment: if the statement is from early in the year and
            the transaction appears to be from a late-month (e.g., December), it may belong
            to the previous year and is adjusted accordingly.
            """
            has_year = bool(re.search(DateFormats.YYYY, tx.date))
            needs_year = not has_year and "y" not in date_format.lower()
            fmt = date_format
            parsed_date = None

            if needs_year:
                tx.date = f"{tx.date} {statement_date.year}"
                fmt += " %Y"

            try:
                parsed_date = datetime.strptime(tx.date, fmt).astimezone()
            except ValueError:
                logger.debug("strptime failed for %s with format %s", tx.date, fmt)
                from dateparser import parse

                parsed_date = parse(tx.date, settings=date_order.settings)

            if not parsed_date:
                msg = f"Could not convert date: {tx.date}"
                raise RuntimeError(msg)

            # Detect cross-year case: e.g., statement is from Jan/Feb, but tx is Dec
            is_cross_year = statement_date.month in START_OF_YEAR_MONTHS and parsed_date.month > YEAR_CUTOFF_MONTH

            if is_cross_year and needs_year:
                parsed_date = parsed_date.replace(year=parsed_date.year - 1)

            return parsed_date.date().isoformat()

        logger.debug("Transforming dates to ISO 8601")

        for tx in transactions:
            tx.date = convert_date(tx)

        return transactions

    @staticmethod
    def load(
        transactions: list[Transaction],
        statement: BaseStatement,
        output_directory: Path | str,
        *,
        preserve_filename: bool,
    ):
        output_directory = Path(output_directory)

        if preserve_filename and statement.file_path:
            filename = f"{Path(statement.file_path).stem}.csv"
        else:
            filename = generate_name(
                statement=statement,
                format_type="file",
                bank_name=statement.bank_name,
                statement_type=statement.statement_type,
                statement_date=statement.statement_date,
            )

        output_path = output_directory / filename
        logger.debug("Writing CSV to file path: %s", output_path)

        with open(output_path, mode="w", encoding="utf8") as file:
            writer = csv.writer(file)

            # header
            writer.writerow(statement.columns)

            for transaction in transactions:
                writer.writerow(
                    [
                        transaction.date,
                        transaction.description,
                        transaction.amount,
                    ]
                )

        return output_path



import logging
import re
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import ClassVar

from dateparser import parse

from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import Columns, SharedPatterns
from monopoly.constants.date import ISO8601
from monopoly.pdf import PdfPage
from monopoly.statements.transaction import (
    Transaction,
    TransactionGroupDict,
    TransactionMatch,
)

# pylint: disable=bad-classmethod-argument
logger = logging.getLogger(__name__)

MIN_BREAK_GAP = 20  # Minimum gap between words and numbers to trigger break


@dataclass
class MatchContext:
    line: str
    lines: list[str]
    idx: int
    description: str
    multiline_config: MultilineConfig | None = None


class DescriptionExtractor:
    """Handles extraction and combination of multiline descriptions."""

    def __init__(self, pattern: re.Pattern):
        self.pattern = pattern

    def get_multiline_description(self, context: MatchContext) -> str | None:
        """Combine a transaction description spanning multiple lines into a single string."""
        if not context.multiline_config:
            return None

        builder = DescriptionBuilder(
            initial_description=context.description,
            description_pos=context.line.find(context.description),
            pattern=self.pattern,
            description_margin=context.multiline_config.description_margin,
        )

        # Handle previous line if within margin
        margin = context.multiline_config.include_prev_margin
        if margin and context.idx > 0:
            builder.include_previous_line(context.lines[context.idx - 1], margin)

        # Include subsequent lines until a break condition is met
        for next_line in context.lines[context.idx + 1 :]:
            if builder.should_break_at_line(next_line):
                break
            builder.append_line(next_line)

        return builder.get_result()


class DescriptionBuilder:
    """Handles the building and validation of multiline descriptions."""

    def __init__(
        self,
        initial_description: str,
        description_pos: int,
        pattern: re.Pattern,
        description_margin: int,
    ):
        self.description = initial_description
        self.description_pos = description_pos
        self.pattern = pattern
        self.words_pattern = re.compile(r"\s[A-Za-z]+")
        self.numbers_pattern = re.compile(SharedPatterns.AMOUNT)
        self.description_margin = description_margin

    @staticmethod
    def get_start_pos(line: str) -> int:
        """Return the starting position of the first word in a line."""
        stripped = line.strip()
        return line.find(stripped.split(" ")[0]) if stripped else -1

    @staticmethod
    def is_within_margin(pos1: int, pos2: int, margin: int) -> bool:
        """Check if two positions are within a specified margin."""
        return abs(pos1 - pos2) <= margin

    def should_break_at_line(self, line: str) -> bool:
        """Determine if processing should stop at the current line."""
        # Blank line
        if not line.strip():
            return True

        # Transaction line
        if self.pattern.search(line):
            return True

        next_pos = self.get_start_pos(line)
        if next_pos and not self.is_within_margin(self.description_pos, next_pos, self.description_margin):
            return True

        words_match = self.words_pattern.search(line)
        numbers_match = self.numbers_pattern.search(line)

        # Exclude footer lines
        if words_match and numbers_match:
            words_end = words_match.span()[1]
            numbers_start = numbers_match.span()[0]
            return numbers_start > words_end and numbers_start - words_end > MIN_BREAK_GAP

        return False

    def include_previous_line(self, prev_line: str, margin: int) -> None:
        """Attempt to include the previous line in the description."""
        prev_line = prev_line.strip()
        if not prev_line or self.pattern.search(prev_line):
            return

        prev_pos = self.get_start_pos(prev_line)
        if self.is_within_margin(self.description_pos, prev_pos, margin):
            self.description = f"{prev_line} {self.description}"

    def append_line(self, line: str) -> None:
        """Append a new line to the current description."""
        self.description += f" {line.strip()}"

    def get_result(self) -> str:
        """Return the final combined description."""
        return self.description


class BaseStatement:
    """
    A dataclass representation of a bank statement.

    Contains PDF pages (their raw text representation in a `list`), and specific bank config.
    """

    statement_type = "base"
    columns: ClassVar[list[str]] = [
        Columns.DATE,
        Columns.DESCRIPTION,
        Columns.AMOUNT,
    ]

    def __init__(
        self,
        pages: list[PdfPage],
        bank_name: str,
        config: StatementConfig,
        header: str,
        file_path: Path | None = None,
    ):
        self.config = config
        self.bank_name = bank_name
        self.pages = pages
        self.header = header
        self.file_path = file_path

    @cached_property
    def number_pattern(self) -> re.Pattern:
        return re.compile(r"[\d.,]+")

    @cached_property
    def decimal_pattern(self) -> re.Pattern:
        return re.compile(r"\d+\.\d+$")

    @cached_property
    def subtotal_pattern(self) -> re.Pattern:
        return re.compile(rf"(?:sub\stotal.*?)\s+{SharedPatterns.AMOUNT}", re.IGNORECASE)

    @property
    def pattern(self):
        pattern = self.config.transaction_pattern
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        return pattern

    def get_transactions(self) -> list[Transaction] | None:
        transactions: list[Transaction] = []

        for page_num, page in enumerate(self.pages):
            for line_num, line in enumerate(page.lines):
                if match := self.pattern.search(line):
                    if self._check_bound(match):
                        continue

                    groupdict = TransactionGroupDict(**match.groupdict())
                    groupdict = self.pre_process_transaction_groupdict(groupdict)

                    transaction_match = TransactionMatch(groupdict, match, page_number=page_num)
                    match = self.pre_process_match(transaction_match)
                    context = MatchContext(
                        line=line,
                        lines=page.lines,
                        idx=line_num,
                        description=match.groupdict.description,
                        multiline_config=self.config.multiline_config,
                    )
                    processed_match = self.process_match(match, context)
                    transaction = Transaction(
                        **processed_match.groupdict,
                        auto_polarity=self.config.transaction_auto_polarity,
                    )
                    transactions.append(transaction)

        if not transactions:
            return None

        return self.post_process_transactions(transactions)

    def _check_bound(self, match: re.Match):
        if (bound := self.config.transaction_bound) and match.span(Columns.AMOUNT)[0] >= bound:
            logger.debug("Transaction exists beyond boundary, ignoring")
            return True
        return False

    def pre_process_transaction_groupdict(self, groupdict: TransactionGroupDict) -> TransactionGroupDict:
        if self.config.multiline_config.multiline_transaction_date:
            if groupdict.transaction_date:
                self.previous_transaction_date = groupdict.transaction_date
            else:
                groupdict.transaction_date = self.previous_transaction_date

        return groupdict

    def pre_process_match(self, transaction_match: TransactionMatch) -> TransactionMatch:
        return transaction_match

    def post_process_transactions(self, transactions: list[Transaction]) -> list[Transaction]:
        return transactions

    def process_match(self, match: TransactionMatch, context: MatchContext) -> TransactionMatch:
        # early exit if no multiline config
        if not context.multiline_config:
            return match

        # early exit if end of page
        if context.idx > len(context.lines) - 1:
            return match

        if context.multiline_config.multiline_polarity:
            match.groupdict.polarity = self.get_multiline_polarity(context)

        if context.multiline_config.multiline_descriptions:
            multiline_description = self.get_multiline_description(context)
            match.groupdict.description = multiline_description
        return match

    def get_multiline_polarity(self, context: MatchContext):
        """
        Pull polarity from the next line, if it can be found on the next line.

        e.g.
        Date                           Description                            Amount
        ----------------------------------------------------------------------------
        24.02.25                       PAYMENT RECEIVED - THANK YOU            79.99
                                                                                  CR

        In this case, the polarity will resolve to CR.
        """
        if re.match(SharedPatterns.POLARITY, context.lines[context.idx + 1]):
            return context.lines[context.idx + 1].strip()
        return None

    def get_multiline_description(
        self,
        context: MatchContext,
    ) -> str:
        """Combine a transaction description spanning multiple lines into a single string."""
        extractor = DescriptionExtractor(self.pattern)
        return extractor.get_multiline_description(context) or ""

    @property
    def failed_safety_message(self) -> str:
        return "Safety check failed - transactions may be inaccurate"

    @cached_property
    def transactions(self):
        return self.get_transactions()

    @cached_property
    def statement_date(self) -> datetime:
        pattern = self.config.statement_date_pattern
        allowed_patterns = (re.Pattern, ISO8601)

        if not isinstance(pattern, allowed_patterns):
            msg = f"Pattern must be one of {allowed_patterns}, not {type(pattern)}"
            raise TypeError(msg)

        for page in self.pages:
            lines = page.lines

            for i, line in enumerate(lines):
                text = self._get_search_text(lines, i, line)

                if match := pattern.search(text):
                    date_string = self._construct_date_string(match)
                    statement_date = parse(
                        date_string=date_string,
                        settings=self.config.statement_date_order.settings,
                    )
                    if statement_date:
                        return statement_date
                    logger.info("Unable to parse statement date %s", date_string)

        # Fallback: Try extracting date from filename
        if filename_date := self._extract_date_from_filename():
            logger.info("Statement date extracted from filename: %s", filename_date)
            return filename_date

        msg = "Statement date not found"
        raise ValueError(msg)

    @staticmethod
    def _construct_date_string(match: re.Match):
        """Construct date with named groups 'day', 'month', and 'year' if they exist, otherwise use group 1."""
        if {"day", "month", "year"}.issubset(match.groupdict()):
            day = match.group("day")
            month = match.group("month")
            year = match.group("year")

            return f"{day}-{month}-{year}"
        return match.group(1)

    def _extract_date_from_filename(self) -> datetime | None:
        """
        Extract statement month and year from the filename.

        Only enabled if the bank config has a filename_fallback_pattern set.

        Supports patterns like:
        - eStatement_Nov2025_2025-11-07T14_50_26.pdf
        - CardStatement_Oct2025.pdf

        Returns datetime object with day set to 1, or None if pattern not found.
        """
        if not self.file_path or not self.config.filename_fallback_pattern:
            return None

        filename = self.file_path.name

        if match := self.config.filename_fallback_pattern.search(filename):
            month_abbr = match.group(1)
            year = match.group(2)
            date_string = f"1 {month_abbr} {year}"

            try:
                return parse(
                    date_string=date_string,
                    settings=self.config.statement_date_order.settings,
                )
            except (ValueError, TypeError) as e:
                logger.warning("Failed to parse date from filename '%s': %s", filename, e)
                return None

        return None

    def _get_search_text(self, lines, i, line):
        """Get text to search, optionally combining multiple lines and removing whitespace."""
        if not self.config.multiline_config:
            return line

        if self.config.multiline_config.multiline_statement_date:
            return " ".join(" ".join(lines[i : i + 3]).split())

        return line

    def get_decimal_numbers(self, lines: list[str]) -> set[float]:
        """
        Return all decimal numbers in a statement.

        This is used to perform a safety check, to make sure no transactions have been missed.
        """
        numbers: set[str] = set()
        for line in lines:
            numbers.update(self.number_pattern.findall(line))
            numbers = {number.replace(",", "") for number in numbers}
            decimal_numbers = {float(number) for number in numbers if self.decimal_pattern.match(number)}
        return decimal_numbers

    def perform_safety_check(self) -> bool:
        """Mandate the perform_safety_check method, which should exist in any child class of Statement."""
        msg = "Subclasses must implement perform_safety_check method"
        raise NotImplementedError(msg)

    def get_all_numbers_from_document(self) -> set:
        """
        Iterate over each page in a statement, and retrieves all decimal numbers in each page.

        This is used by child classes with implementations of perform_safety_check().
        """
        numbers = set()

        for page in self.pages:
            lines = page.lines
            numbers.update(self.get_decimal_numbers(lines))
        numbers.add(self.get_subtotal_sum())
        return numbers

    def get_subtotal_sum(self):
        """
        Retrieve the subtotals from a document, and calculates the total.

        Useful for statements that don't give a total figure over
        several cards/months in a single statement.
        """
        subtotals: list[str] = []
        for page in self.pages:
            subtotals.extend(
                match.groupdict()[Columns.AMOUNT]
                for line in page.lines
                if (match := self.subtotal_pattern.search(line))
            )
        cleaned_subtotals = [float(amount.replace(",", "")) for amount in subtotals]
        return sum(cleaned_subtotals)


class SafetyCheckError(Exception):
    """Raised if safety check fails."""



import logging
import re
from collections import defaultdict
from functools import cached_property

from monopoly.constants import EntryType, SharedPatterns
from monopoly.pdf import MetadataIdentifier, PdfPage

from .patterns import DateMatch, DatePattern, PatternMatcher

logger = logging.getLogger(__name__)


DEBIT_MATCH_THRESHOLD = 1.1  # Avg matches per line above this are considered debit
EXPECTED_DATE_COUNT_PER_LINE = 2  # One transaction date + one posting date
EXPECTED_DATE_SPAN_COUNT = 2
MULTILINE_DISTANCE_THRESHOLD = 2  # Avg line gap above this is considered multiline


class GenericParserError(Exception):
    """Exception raised when the generic parser fails to find exceptions."""


class DatePatternAnalyzer:
    def __init__(self, pages: list[PdfPage], metadata: MetadataIdentifier):
        self.pages = pages
        self.matcher = PatternMatcher(pages)
        self.pattern: DatePattern = self.matcher.get_transaction_pattern()
        self.matches = self.pattern.matches
        self.spans = self.matcher.get_transaction_spans(self.pattern)
        self.metadata = metadata

    @cached_property
    def lines_before_first_transaction(self) -> list[str]:
        page_num, line_num = self.get_first_transaction_location()

        page = self.pages[page_num]

        # calculate the starting index, ensuring it's not negative
        start_index = max(line_num - 20, 0)
        lines_before_transaction = page.lines[start_index:line_num]
        return list(reversed(lines_before_transaction))

    def get_items_with_same_page_and_line(self) -> list[tuple]:
        """Group date items with the same line and page number."""
        grouped_items: defaultdict[str, list] = defaultdict(list)
        for match in self.matches:
            key = f"{match.page_number}-{match.line_number}"
            grouped_items[key].append(match.parsed_date)

        # Convert the dictionary into a list of tuple
        return [(key, *values) for key, values in grouped_items.items()]

    def is_transaction_date_first(self) -> bool:
        """Check that in cases where date_1 > date_2, the first date is identified as the posting date."""
        # only focus on our main transaction pattern
        lines_with_multiple_dates = self.get_items_with_same_page_and_line()
        for line in lines_with_multiple_dates:
            logger.debug("Checking dates in line: %s", line)
            if len(line) > EXPECTED_DATE_COUNT_PER_LINE:
                _, date_1, date_2 = line
                if date_1 and date_2 and date_1 != date_2:
                    logger.debug("%s != %s, returning %s", date_1, date_2, date_1 < date_2)
                    return date_1 < date_2

        logger.warning("Could not identify difference between transaction and posting date")
        return True

    def create_transaction_pattern(self) -> re.Pattern:
        """Create a regex pattern that will be used for date parsing by the generic statement handler."""
        logger.debug("Creating date pattern for %s", self.pattern.name)
        transaction_regex = rf"(?P<transaction_date>{self.pattern.regex})\s+"

        # if there are two spans, we attempt to determine which is the
        # posting date, and which is the transaction date
        if len(self.spans) == EXPECTED_DATE_SPAN_COUNT:
            logger.debug("Creating group for posting date")
            posting_regex = rf"(?P<posting_date>{self.pattern.regex})\s+"
            if self.is_transaction_date_first():
                pattern = transaction_regex + posting_regex
            else:
                pattern = posting_regex + transaction_regex
        else:
            pattern = transaction_regex

        pattern += SharedPatterns.DESCRIPTION

        # if the statement is debit-type statement, we add the amount + balance patterns
        if self.get_statement_type() == EntryType.DEBIT:
            pattern += SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
            pattern += SharedPatterns.BALANCE

        if self.get_statement_type() == EntryType.CREDIT:
            pattern += SharedPatterns.AMOUNT_EXTENDED

        return re.compile(pattern, re.IGNORECASE)

    def create_statement_date_pattern(self) -> re.Pattern:
        """Create a regex pattern for the statement date based on the first statement date."""
        statement_date = self.matcher.get_statement_date_pattern()
        return re.compile(f"({statement_date})")

    def get_statement_type(self) -> str:
        """
        Determine whether a statement is a debit or credit statement.

        Since debit and credit statements need to be handled differently,
        we use the average number of matches per line as a way to determine
        if a statement is a debit or credit statement.

        For example, a debit statement transaction line might be something like:
        `01 OCT   ValueVille    123.12     2,000.00`
        where the first number is either a deposit/withdrawal,
        and the next number is a balance.
        """
        # match up to amounts like `123.12     2,000.00` while ignoring words between
        amount_pattern = re.compile(r"(?P<amount>\d{1,3}(,\d{3})*(\.\d+)\d{1,3}(,\d{3})*(\.\d+)?)")

        def get_num_amounts_per_line(lines: list[DateMatch]) -> float:
            matches = [result for line in lines for result in amount_pattern.finditer(line.line)]
            return len(matches) / len(lines)

        average_matches_per_line = get_num_amounts_per_line(self.matches)
        # this technically should be 1.0, but 1.1 provides some
        # slight tolerance for error
        if average_matches_per_line > DEBIT_MATCH_THRESHOLD:
            logger.debug(
                "Found %s amounts per line - assuming debit statement",
                average_matches_per_line,
            )
            return EntryType.DEBIT

        return EntryType.CREDIT

    def get_debit_statement_header_line(self, lines_before_first_transaction) -> str:
        header_pattern = re.compile(r"\b(date.*$)", re.IGNORECASE)

        for line in lines_before_first_transaction:
            if line:
                # assume that the header always starts with `date` or `DATE`
                result = header_pattern.search(line)
                if result:
                    generalized_result = re.sub(r"\s{3,}", r"\\s+", result.string)
                    escaped_result = re.escape(generalized_result)
                    escaped_result = escaped_result.replace("\\\\s\\+", r"\s+")
                    logger.debug("Found header statement: %s", escaped_result)
                    return escaped_result

        msg = "Could not find debit statement header line"
        raise RuntimeError(msg)

    def check_if_multiline(self) -> bool:
        """Check if the statement should be treated as a multiline statement."""
        line_numbers = [item.line_number for item in self.matches]
        line_distance = [line_numbers[i] - line_numbers[i - 1] for i in range(1, len(line_numbers))]
        # remove negative numbers, since this indicates that the transaction is
        # from the next page
        line_distance = [i for i in line_distance if i > 0]

        # if there's more than two lines between each transaction on average
        # we assume that this is a multiline statement
        average_line_distance = sum(line_distance) / len(line_distance)
        logger.debug(
            "Assuming multiline statement: average distance between transaction lines: %s",
            average_line_distance,
        )
        return average_line_distance > MULTILINE_DISTANCE_THRESHOLD

    def create_previous_balance_regex(self) -> re.Pattern | None:
        """
        Check for a previous balance line items.

        Makes the assumption that the previous balance line item, if it
        exists, will be the line before the first line item with a date.

        e.g.
        ```
        PREVIOUS BALANCE                       6,492.54
        05 OCT PAYMENT - DBS INTERNET/WIRELESS 6,492.54 CR
        ```
        """
        page_num, line_num = self.get_first_transaction_location()

        # grab the first five lines before the first transaction
        page = self.pages[page_num]
        lines_before_first_transaction = list(reversed(page.lines[line_num - 5 : line_num]))

        balance_pattern = re.compile(SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT)

        for line in lines_before_first_transaction:
            balance_match = balance_pattern.search(line)
            if balance_match:
                logger.debug("Balance match found")
                if word_match := re.search(r"(\b[a-zA-Z'\s]+)", balance_match.string):
                    words = word_match[0].strip()
                    pattern = SharedPatterns.DESCRIPTION.replace(".*?", words) + SharedPatterns.AMOUNT
                    logger.debug("Found words, generated pattern %s", pattern)
                    return re.compile(pattern)
        return None

    def get_first_transaction_location(self):
        # uses the transaction pattern to find the first transaction
        pattern = self.create_transaction_pattern()

        for page_num, page in enumerate(self.pages):
            for line_num, line in enumerate(page.lines):
                match = pattern.search(line)
                if match:
                    logger.debug("Found first transaction at page %s line %s", page_num, line_num)
                    return page_num, line_num
        msg = f"Could not find any transactions. PDF metadata: {self.metadata})"
        raise GenericParserError(msg)



from datetime import datetime
from pathlib import Path

import pytest
from test_utils.skip import skip_if_encrypted
from test_utils.transactions import get_transactions_as_dict, read_transactions_from_csv

from monopoly.banks import Citibank, Dbs, Hsbc, Maybank, Ocbc, StandardChartered, Trust
from monopoly.banks.base import BankBase
from monopoly.pdf import PdfDocument, PdfParser
from monopoly.pipeline import Pipeline
from monopoly.statements import CreditStatement

test_cases = [
    (Citibank, -1414.07, datetime(2022, 11, 15)),
    (Dbs, -16969.17, datetime(2023, 10, 15)),
    (Hsbc, -1218.2, datetime(2023, 8, 20)),
    (Maybank, -1259.28, datetime(2024, 11, 8)),
    (Ocbc, -702.1, datetime(2023, 8, 1)),
    (StandardChartered, -82.45, datetime(2023, 5, 16)),
    (Trust, -681.27, datetime(2024, 8, 18)),
]


@skip_if_encrypted
@pytest.mark.parametrize(
    "bank, total_amount, statement_date",
    test_cases,
)
def test_bank_credit_statements(
    bank: BankBase,
    total_amount: float,
    statement_date: datetime,
):
    test_directory = Path(__file__).parent / bank.name / "credit"

    document = PdfDocument(test_directory / "input.pdf")
    parser = PdfParser(bank, document)
    pipeline = Pipeline(parser)
    statement: CreditStatement = pipeline.extract()

    # check raw data
    expected_raw_transactions = read_transactions_from_csv(test_directory, "raw.csv")
    raw_transactions_as_dict = get_transactions_as_dict(statement.transactions)
    expected_transaction_total_amount = [transaction.amount for transaction in statement.transactions]
    assert expected_raw_transactions == raw_transactions_as_dict
    assert round(sum(expected_transaction_total_amount), 2) == total_amount
    assert statement.statement_date == statement_date

    # check transformed data
    expected_transformed_transactions = read_transactions_from_csv(test_directory, "transformed.csv")
    transformed_transactions = pipeline.transform(statement)
    transformed_transactions_as_dict = get_transactions_as_dict(transformed_transactions)
    assert expected_transformed_transactions == transformed_transactions_as_dict



import logging
import re

from monopoly.banks.base import BankBase
from monopoly.config import MultilineConfig, StatementConfig
from monopoly.constants import BankNames, CreditTransactionPatterns, EntryType
from monopoly.identifiers import TextIdentifier

logger = logging.getLogger(__name__)


class Trust(BankBase):
    name = BankNames.TRUST

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(
            r"-\s*"
            r"(?P<day>\d{2})\s+"
            r"(?P<month>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+"
            r".*?"
            r"(?P<year>20\d{2}\b)"
        ),
        header_pattern=re.compile(r"(Posting date.*Description.*Amount in SGD)"),
        transaction_pattern=CreditTransactionPatterns.TRUST,
        multiline_config=MultilineConfig(
            multiline_descriptions=True,
            include_prev_margin=99,
            multiline_statement_date=True,
        ),
        safety_check=True,
        transaction_date_format="%d %b",
    )

    identifiers = [[TextIdentifier("Trust Bank Singapore Limited")]]

    statement_configs = [credit]



import csv
from pathlib import Path

from monopoly.statements import Transaction


def read_transactions_from_csv(directory: Path, file_name: str) -> list[dict[str, str]]:
    transactions = []
    with open(directory / file_name, "r") as file:
        reader = csv.DictReader(file)
        for row in reader:
            transactions.append(row)
    return transactions


def get_transactions_as_dict(transactions: list[Transaction]) -> list[dict[str, str]]:
    transactions_as_dict = []
    for transaction in transactions:
        transactions_as_dict.append(transaction.as_raw_dict())
    return transactions_as_dict



import contextvars
import functools
import logging
import sys

file_context = contextvars.ContextVar("file_context", default="N/A")
default_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
verbose_formatter = logging.Formatter("%(asctime)s - [%(file_context)s] - %(name)s - %(levelname)s - %(message)s")


class ContextFilter(logging.Filter):
    """
    Context filter that is used to inject the PDF file name into the logs.

    Only enabled in verbose mode.
    """

    def filter(self, record):
        record.file_context = file_context.get()
        return True


def set_verbose_logging():
    """Switches the logger to a more verbose format and sets the level to DEBUG."""
    logger = logging.getLogger("monopoly")
    logger.setLevel(logging.DEBUG)
    for handler in logger.handlers:
        handler.setFormatter(verbose_formatter)


def get_logger() -> logging.Logger:
    logger = logging.getLogger("monopoly")

    if logger.hasHandlers():
        return logger

    handler = logging.StreamHandler(sys.stdout)

    handler.addFilter(ContextFilter())
    handler.setFormatter(default_formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    logging.getLogger("pdf2john").setLevel(logging.ERROR)
    logging.getLogger("pyhanko").setLevel(logging.ERROR)
    logging.getLogger("tzlocal").setLevel(logging.ERROR)
    logging.getLogger("pikepdf").setLevel(logging.ERROR)
    return logger


def worker_log_setup(*, verbose: bool):
    if verbose:
        set_verbose_logging()


def setup_logs(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("verbose"):
            set_verbose_logging()
        return func(*args, **kwargs)

    return wrapper



import re

from monopoly.banks.base import BankBase
from monopoly.config import DateOrder, MultilineConfig, StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.constants.date import DateFormats
from monopoly.identifiers import MetadataIdentifier, TextIdentifier


class CIBC(BankBase):
    name = BankNames.CIBC

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        header_pattern=re.compile(
            r"(\s+)?Date\s+Description\s+Withdrawals\ \(\$\)\s+Deposits\ \(\$\)\s+Balance\ \(\$\)"
        ),
        statement_date_pattern=re.compile(
            rf"(?:to\s+)?(?P<date>{DateFormats.MMM}\s+{DateFormats.DD},\s+{DateFormats.YYYY})"
        ),
        transaction_pattern=DebitTransactionPatterns.CIBC,
        transaction_date_format="%b %d",
        transaction_date_order=DateOrder("DMY"),
        statement_date_order=DateOrder("DMY"),
        multiline_config=MultilineConfig(
            multiline_transaction_date=True,
            multiline_descriptions=True,
        ),
        safety_check=True,
        transaction_auto_polarity=True,
    )

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=re.compile(
            rf"(?:to\s+)?(?P<date>{DateFormats.MMM}\s+{DateFormats.DD},\s+{DateFormats.YYYY})"
        ),
        header_pattern=re.compile(r"\s+date\s+date\s+Description\s+Spend Categories\s+Amount\(\$\)"),
        prev_balance_pattern=StatementBalancePatterns.CIBC,
        transaction_pattern=CreditTransactionPatterns.CIBC,
        transaction_date_format="%b %d",
        transaction_auto_polarity=False,
    )

    identifiers = [
        [
            TextIdentifier(text="CIBC Account Statement"),
            MetadataIdentifier(producer="iText® 5.5.13.2 ©2000-2020 iText Group NV (AGPL-version)"),
        ],
        [
            MetadataIdentifier(
                author="CIBC", producer="Ricoh Americas Corporation, AFP2PDF Plus Version: 1.300.71, Linux"
            )
        ],
    ]

    statement_configs = [debit, credit]



# monopoly/detectors/bank_detector.py

import logging
from typing import TYPE_CHECKING

from monopoly.identifiers import Identifier  # Import base class
from monopoly.pdf import PdfDocument

if TYPE_CHECKING:
    from monopoly.banks.base import BankBase

logger = logging.getLogger(__name__)


class BankDetector:
    def __init__(self, document: PdfDocument):
        self.document = document
        self.metadata_identifier = document.metadata_identifier

    def detect_bank(self, banks: list[type["BankBase"]]) -> type["BankBase"] | None:
        """Detect the bank by checking its identifier groups against the document."""
        logger.debug("Found PDF properties: %s", self.metadata_identifier)
        for bank in banks:
            # A bank is identified if ANY of its identifier groups is a full match
            if any(self.identifiers_match(group) for group in bank.identifiers):
                logger.debug("Identified statement bank: %s", bank.__name__)
                return bank
        return None

    def identifiers_match(self, identifiers: list[Identifier]) -> bool:
        """
        Check if ALL identifiers in a given group match the document.

        An empty group of identifiers does not constitute a match.
        """
        if not identifiers:
            return False

        # Polymorphically call .matches() on each identifier in the group.
        # ALL of them must return True for the group to match.
        return all(identifier.matches(self) for identifier in identifiers)



from datetime import datetime
import re

from monopoly.config import MultilineConfig
from monopoly.statements import BaseStatement


def test_statement_date_multiline(statement: BaseStatement):
    statement.pages[0].lines = [
        "   01-01                                                      23 Apr 2025 - 23 May",
        "                                                                          Statement cycle",
        "   02-123                                                                            2025",
    ]

    statement.config.statement_date_pattern = re.compile(
        r"-\s*"
        r"(?P<day>\d{2})\s+"
        r"(?P<month>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+"
        r".*?"
        r"(?P<year>20\d{2}\b)"
    )
    statement.config.multiline_config = MultilineConfig(multiline_statement_date=True)
    assert statement.statement_date == datetime(2025, 5, 23)



from unittest.mock import PropertyMock, patch

import pytest

from monopoly.banks.base import BankBase
from monopoly.banks.detector import BankDetector
from monopoly.identifiers import MetadataIdentifier, TextIdentifier
from monopoly.pdf import PdfDocument


@pytest.fixture
def mock_encrypted_document():
    with patch("monopoly.pdf.PdfParser.document", new_callable=PropertyMock) as mock_document:
        mock_document_instance = mock_document.return_value
        type(mock_document_instance).metadata = PropertyMock(return_value={})
        yield mock_document


class MockBankOne(BankBase):
    name = "bank1"
    statement_configs = None
    debit = None
    credit = None
    identifiers = [
        [
            MetadataIdentifier(creator="foo", producer="bar"),
        ]
    ]


class MockBankTwo(BankBase):
    name = "bank2"
    statement_configs = None
    debit = None
    credit = None
    identifiers = [[MetadataIdentifier(creator="Adobe Acrobat 23.3", producer="Adobe Acrobat Pro (64-bit)")]]


class MockBankThree(BankBase):
    name = "bank3"
    statement_configs = None
    debit = None
    credit = None
    identifiers = [[MetadataIdentifier(creator="asdasd", producer="qwerty")]]


class MockBankWithMultipleTextIdentifier(BankBase):
    name = "bank-multi"
    statement_configs = None
    debit = None
    credit = None
    identifiers = [
        [
            MetadataIdentifier(creator="foo", producer="bar"),
            TextIdentifier("specific_string"),
            TextIdentifier("other_specific_string"),
        ]
    ]


class MockBankWithOnlyTextIdentifier(BankBase):
    name = "bank-text"
    statement_configs = None
    debit = None
    credit = None
    identifiers = [
        [
            TextIdentifier("foo"),
            TextIdentifier("baz"),
            TextIdentifier("bar"),
        ]
    ]


unencrypted_file_path = "path/to/unencrypted.pdf"
encrypted_file_path = "path/to/encrypted.pdf"


def test_auto_detect_bank_identified(metadata_analyzer: BankDetector):
    metadata_analyzer.metadata_identifier = MetadataIdentifier(
        creator="Adobe Acrobat 23.3", producer="Adobe Acrobat Pro (64-bit)"
    )

    mock_banks_list = [MockBankOne, MockBankTwo]

    bank = metadata_analyzer.detect_bank(mock_banks_list)
    assert bank.__name__ == MockBankTwo.__name__


def test_detect_bank_not_identified(metadata_analyzer: BankDetector):
    metadata_analyzer.metadata_identifier = MetadataIdentifier(creator="asdf", producer="qwerty")
    mock_banks_list = [MockBankThree]
    assert not metadata_analyzer.detect_bank(mock_banks_list)


@patch.object(PdfDocument, "raw_text", new_callable=PropertyMock)
def test_detect_bank_with_text_identifier(mock_raw_text, metadata_analyzer: BankDetector):
    mock_raw_text.return_value = "specific_string, other_specific_string"
    metadata_analyzer.metadata_identifier = MetadataIdentifier(creator="foo", producer="bar")

    mock_banks_list = [MockBankTwo, MockBankWithMultipleTextIdentifier]
    bank = metadata_analyzer.detect_bank(mock_banks_list)

    assert bank.__name__ == MockBankWithMultipleTextIdentifier.__name__


@patch.object(PdfDocument, "raw_text", new_callable=PropertyMock)
def test_detect_bank_with_not_matching_text_identifier(mock_raw_text, monkeypatch, metadata_analyzer: BankDetector):
    mock_raw_text.return_value = "not_a_match"
    metadata_analyzer.metadata_identifier = MetadataIdentifier(creator="foo", producer="bar")

    mock_banks_list = [MockBankTwo, MockBankWithMultipleTextIdentifier]
    monkeypatch.setattr("monopoly.banks.banks", mock_banks_list)

    assert not metadata_analyzer.detect_bank(mock_banks_list)


@patch.object(PdfDocument, "raw_text", new_callable=PropertyMock)
def test_detect_bank_with_only_text_identifier(mock_raw_text, metadata_analyzer: BankDetector):
    mock_raw_text.return_value = "foo baz bar"
    metadata_analyzer.metadata_identifier = MetadataIdentifier(creator="foo", producer="bar")

    mock_banks_list = [
        MockBankWithMultipleTextIdentifier,
        MockBankWithOnlyTextIdentifier,
    ]

    bank = metadata_analyzer.detect_bank(mock_banks_list)
    assert bank.__name__ == MockBankWithOnlyTextIdentifier.__name__



import os
import re
from pathlib import Path
import pytest
from click.testing import CliRunner
from monopoly.cli.models import Report, Result
from monopoly.cli import (
    get_statement_paths,
    monopoly,
    pprint_transactions,
)
from monopoly.statements.transaction import Transaction


@pytest.fixture
def mock_results():
    result1 = Result(source_file_name="statement1.pdf", target_file_name="processed1.csv")
    result2 = Result(
        source_file_name="statement2.pdf",
        error_info={
            "message": "Error",
            "traceback": "Traceback (most recent call last)",
        },
    )
    return [result1, result2]


@pytest.fixture
def test_directory() -> Path:
    return Path("tests/integration/test_cli").resolve()


@pytest.fixture
def cli_runner():
    yield CliRunner()


def test_display_report(mock_results, capsys):
    report = Report(results=mock_results)
    report.display_report()
    captured = capsys.readouterr()
    printed_output = captured.out.strip().split("\n")
    assert "1 statement(s) had errors while processing" in printed_output
    assert "1 statement(s) processed" in printed_output
    assert "statement1.pdf -> processed1.csv" in printed_output
    assert "statement2.pdf -- Error" in printed_output

    # Test verbose output
    report.display_report(verbose=True)
    captured = capsys.readouterr()
    printed_output = captured.out.strip().split("\n")
    assert "statement2.pdf -- Traceback (most recent call last)" in printed_output


def test_help_command(cli_runner: CliRunner) -> None:
    help_results = cli_runner.invoke(monopoly, args="--help")
    assert help_results.exit_code == 0
    assert help_results.stdout.startswith("Usage: monopoly")


def test_monopoly_output(cli_runner: CliRunner, tmp_path: Path):
    output_dir = tmp_path / "results"
    output_dir.mkdir()

    # Pass the PDF path from the fixture and the temporary output directory
    result = cli_runner.invoke(monopoly, ["src/monopoly/examples/example_statement.pdf", "--output", str(output_dir)])

    assert result.exit_code == 0
    assert "1 statement(s) processed" in result.output

    # Check that the output file was created with the correct name
    expected_csv_path = output_dir / "example-credit-2023-07-74498f.csv"
    assert expected_csv_path.exists()
    assert str(expected_csv_path.name) in result.output


def test_monopoly_output_preserve_filename(cli_runner: CliRunner, tmp_path: Path):
    output_dir = tmp_path / "results"
    output_dir.mkdir()

    result = cli_runner.invoke(
        monopoly,
        [
            "src/monopoly/examples/example_statement.pdf",
            "--output",
            str(output_dir),
            "--preserve-filename",
        ],
    )

    assert result.exit_code == 0
    assert "1 statement(s) processed" in result.output

    expected_csv_path = output_dir / "example_statement.csv"
    assert expected_csv_path.exists()
    assert str(expected_csv_path.name) in result.output


def test_monopoly_preserve_filename_without_output(cli_runner: CliRunner):
    with cli_runner.isolated_filesystem():
        repo_root = Path(__file__).resolve().parents[2]
        src_pdf = repo_root / "src/monopoly/examples/example_statement.pdf"
        Path("example_statement.pdf").write_bytes(src_pdf.read_bytes())

        result = cli_runner.invoke(
            monopoly,
            [
                "example_statement.pdf",
                "--preserve-filename",
            ],
        )

        assert result.exit_code == 0
        expected_csv_path = Path("example_statement.csv")
        assert expected_csv_path.exists()
        assert str(expected_csv_path.name) in result.output


def test_monopoly_pprint(cli_runner: CliRunner):
    result = cli_runner.invoke(
        monopoly, ["src/monopoly/examples/example_statement.pdf", "--pprint", "--single-process"]
    )

    assert result.exit_code == 0

    output = result.output
    assert "| date       | description                       |   amount |" in output
    assert "| 2023-07-02 | LAST MONTH'S BALANCE              |  -412.16 |" in output
    assert "| 2023-07-18 | CASH REBATE                       |     1.38 |" in output


def test_monopoly_no_pdf(cli_runner: CliRunner):
    with cli_runner.isolated_filesystem():
        with open("file.txt", "w") as f:
            f.write("not a pdf file")
        result = cli_runner.invoke(monopoly, ["file.txt"])
    assert result.exit_code == 1
    assert "Could not find .pdf files" in result.output


def test_get_statement_paths(test_directory: Path):
    path = test_directory
    expected = {
        path / "top_level.pdf",
        path / "top_level_2.pdf",
        path / "nested_directory/nested.pdf",
    }
    res = get_statement_paths(test_directory.iterdir())
    assert res == expected


def test_version_command(cli_runner: CliRunner):
    results = cli_runner.invoke(monopoly, args="--version")
    assert results.exit_code == 0
    assert results.stdout.startswith("monopoly, version ")
    semver_pattern = r"\d+\.\d+.\d+"
    match = re.search(semver_pattern, results.stdout)
    assert match, "Semantic version number not in output"


def test_pprint_transactions(capsys, statement):
    file = Path("test_file.md")
    transactions = [
        Transaction(transaction_date="2023-01-01", description="Transaction 1", amount=100.00),
        Transaction(transaction_date="2023-01-01", description="Transaction 2", amount=123.12),
    ]
    pprint_transactions(transactions, statement, file)
    captured = capsys.readouterr()
    expected_output = (
        "test_file.md\n"
        "+------------+---------------+----------+\n"
        "| date       | description   |   amount |\n"
        "|------------+---------------+----------|\n"
        "| 2023-01-01 | Transaction 1 |     -100 |\n"
        "| 2023-01-01 | Transaction 2 |  -123.12 |\n"
        "+------------+---------------+----------+\n"
        "\n"
    )
    assert captured.out == expected_output
